@@ -181,19 +181,6 @@
 //!
 //! assert_eq!(results, vec![Ok(0), Ok(1), Ok(2), Err(()), Ok(4)])
 //! ```
-//! ## Overhead Benchmark
-//! Platform: Macbook Air (2015 Late) 8 GB RAM, Intel Core i5, 1.6GHZ (2 Core).
-//!
-//! ### Result
-//! One million (1,000,000) empty iteration for each run.
-//! ```text
-//! test iter_async::test_par_iter_async::bench_into_par_iter_async ... bench: 120,003,398 ns/iter (+/- 52,401,527)
-//! test test_par_iter::bench_into_par_iter_sync                    ... bench:  98,472,767 ns/iter (+/- 9,901,593)
-//! ```
-//!
-//! Result:
-//! - Async iterator overhead `120,003,398 / 1,000,000 = 120 ns (+/- 52 ns)`.
-//! - Sync iterator overhead  ` 98,472,767 / 1,000,000 =  98 ns (+/- 10 ns)`.
 //!
 //! ## Implementation Note
 //!
@@ -203,33 +190,37 @@
 //!   get blocked. The channel size is hard-coded to 100 for each thread.
 //! - The number of threads equals to the number of logical cores.
 //!
-//! ### Synchronization and Exception Handling
-//! - When each thread fetch a task, it registers its thread ID and task ID into a registry.
-//! - When `next()` is called, the consumer fetch from the task registry the next thread ID.
-//! - `next()` returns None if there is no more task or if some Error occurs.
+//! ### Synchronization Mechanism
+//! - When each thread fetch a task, it registers its thread ID (`thread_number`)
+//!   and the task ID (`task_number`) into a mpsc channel.
+//! - When `next()` is called, the consumer fetch from the task registry
+//!   (`task_order`) the next thread ID and task ID.
+//! - If `next()` detect that some thread has not produced result due to exception,
+//!   it calls `kill()`, which stop threads from fetching new tasks,
+//!   flush remaining tasks, and joining the worker threads.
+//!
+//! ### Error handling and Dropping
+//! - When any exception occurs, stop producers from fetching new task.
+//! - Before dropping the structure, stop all producers from fetching tasks,
+//!   flush all remaining tasks, and join all threads..
 //!
 mod iter_async;
 
-use crossbeam::channel::{bounded, Receiver};
+use crossbeam::channel::{bounded, unbounded, Receiver, Sender};
 pub use iter_async::*;
 use num_cpus;
-use std::ops::Deref;
-use std::sync::atomic::{AtomicBool, AtomicIsize, Ordering};
-use std::sync::Arc;
+use std::sync::atomic::{AtomicBool, Ordering};
+use std::sync::{Arc, Mutex};
 use std::thread;
 use std::thread::JoinHandle;
-use std::time::Duration;
 
 const MAX_SIZE_FOR_THREAD: usize = 100;
 
-///
-/// lock-free sequential parallel iterator
-///
 pub trait IntoParallelIteratorSync<R, T, TL, F>
 where
     F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
-    T: Send + 'static,
-    TL: Send + IntoIterator<Item = T> + 'static,
+    T: Send,
+    TL: Send + IntoIterator<Item = T>,
     <TL as IntoIterator>::IntoIter: Send + 'static,
     R: Send,
 {
@@ -282,8 +273,8 @@ where
 impl<R, T, TL, F> IntoParallelIteratorSync<R, T, TL, F> for TL
 where
     F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
-    T: Send + 'static,
-    TL: Send + IntoIterator<Item = T> + 'static,
+    T: Send,
+    TL: Send + IntoIterator<Item = T>,
     <TL as IntoIterator>::IntoIter: Send + 'static,
     R: Send + 'static,
 {
@@ -292,121 +283,18 @@ where
     }
 }
 
-///
-/// A lookup table to register and look up corresponding thread id for a task.
-///
-/// `-1` represents the task ID is not yet registered
-/// or the task does not exist.
-///
-#[derive(Clone)]
-struct TaskRegistry(
-    // vector of thread ID
-    Arc<Vec<AtomicIsize>>
-);
-
-impl Deref for TaskRegistry {
-    type Target = Vec<AtomicIsize>;
-
-    fn deref(&self) -> &Self::Target {
-        self.0.deref()
-    }
-}
-
-impl TaskRegistry {
-
-    ///
-    /// Initialize the registry with `-1` to represent an empty registry
-    ///
-    /// The `size` must be (just) big enough to ensure that no key collision would
-    /// possibly occur.
-    ///
-    fn new(size: usize) -> TaskRegistry {
-        TaskRegistry(Arc::new((0..size).map(|_| AtomicIsize::new(-1)).collect()))
-    }
-
-    ///
-    /// Look up a thread_number of a task and set that slot to `-1`.
-    ///
-    /// This function blocks to wait for a task to be registered,
-    /// unless all worker threads have stopped so that no more new
-    /// task can possibly be registered.
-    ///
-    /// It should block very rarely since task dispatcher is not blocking,
-    /// and is registered immediately after fetching in `get_task`.
-    ///
-    /// returns `None` only when all worker threads have stopped
-    ///
-    #[inline(always)]
-    pub(crate) fn lookup(&self, task_id: usize) -> Option<isize> {
-        let registry_len = self.len();
-        let pos = Self::id_to_key(task_id, registry_len);
-        loop {
-            // check if worker threads are still active
-            if Arc::strong_count(&self.0) > 1 {
-                let thread_num = self[pos].swap(-1, Ordering::SeqCst);
-                // if `-1` is read, would continue in the loop
-                if thread_num >= 0 {
-                    return Some(thread_num);
-                } else {
-                    // a short sleep before checking again
-                    thread::sleep(Duration::from_nanos(500))
-                }
-            // if worker threads are no more active, might return `None`
-            } else {
-                let thread_num = self[pos].swap(-1, Ordering::SeqCst);
-                return if thread_num >= 0 {
-                    Some(thread_num)
-                } else {
-                    None
-                };
-            }
-        }
-    }
-
-    ///
-    /// When a worker fetches a new task, it calls `register` to tell
-    /// the user thread its own thread ID.
-    ///
-    /// register the worker thread number of a task
-    ///
-    #[inline(always)]
-    pub(crate) fn register(&self, task_id: usize, thread_id: isize) {
-        let registry_len = self.len();
-        let key = Self::id_to_key(task_id, registry_len);
-        // never overwrite
-        debug_assert_eq!(self[key].load(Ordering::SeqCst), -1);
-        self[key].store(thread_id, Ordering::SeqCst);
-    }
-
-    /// key of task ID in registry
-    #[inline(always)]
-    fn id_to_key(task_id: usize, registry_len: usize) -> usize {
-        task_id % registry_len
-    }
-}
-
-///
-/// implementation of lock-free sequential parallel iterator
-///
+/// iterate through blocks according to array index.
 pub struct ParIterSync<R> {
-
     /// Result receivers, one for each worker thread
-    output_receivers: Vec<Receiver<R>>,
-
-    /// Lookup table to register worker thread number corresponding to tasks
-    task_registry: TaskRegistry,
-
+    receivers: Vec<Receiver<R>>,
+    /// Receiver<thread>
+    task_order: Receiver<usize>,
     /// handles to join worker threads
     worker_thread: Option<Vec<JoinHandle<()>>>,
-
-    /// atomic flag to stop workers from fetching new tasks
+    /// flag to stop workers from fetching new tasks
     iterator_stopper: Arc<AtomicBool>,
-
-    /// if this is `true`, it must guarantee that all worker threads have stopped
+    /// indicate that workers have all been killed
     is_killed: bool,
-
-    /// current task id
-    current: usize,
 }
 
 impl<R> ParIterSync<R>
@@ -419,54 +307,38 @@ where
     pub fn new<T, TL, F>(tasks: TL, task_executor: F) -> Self
     where
         F: Send + Clone + 'static + Fn(T) -> Result<R, ()>,
-        T: Send + 'static,
-        TL: Send + IntoIterator<Item = T> + 'static,
+        T: Send,
+        TL: Send + IntoIterator<Item = T>,
         <TL as IntoIterator>::IntoIter: Send + 'static,
     {
         let cpus = num_cpus::get();
         let iterator_stopper = Arc::new(AtomicBool::new(false));
-
-        // `(1 + MAX_SIZE_FOR_THREAD)` * cpus as there might be one more fetching after send blocking
-        let task_registry: TaskRegistry = TaskRegistry::new((1 + MAX_SIZE_FOR_THREAD) * cpus);
-
-        // this thread dispatches tasks to worker threads
-        let (dispatcher, task_receiver) = bounded(MAX_SIZE_FOR_THREAD * cpus);
-        let sender_thread = thread::spawn(move || {
-            for (task_id, t) in tasks.into_iter().enumerate() {
-                if dispatcher.send((t, task_id)).is_err() {
-                    break;
-                }
-            }
-        });
-
-        // spawn worker threads
-        let mut handles = Vec::with_capacity(cpus + 1);
-        let mut output_receivers = Vec::with_capacity(cpus);
-        for thread_number in 0..cpus as isize {
-            let (output_sender, output_receiver) = bounded(MAX_SIZE_FOR_THREAD);
-            let task_receiver = task_receiver.clone();
-            let task_registry = task_registry.clone();
+        // worker master
+        let (task_register, task_order) = unbounded();
+        let tasks = Arc::new(Mutex::new(tasks.into_iter()));
+        let mut handles = Vec::with_capacity(cpus);
+        let mut receivers = Vec::with_capacity(cpus);
+        for thread_number in 0..cpus {
+            let (sender, receiver) = bounded(MAX_SIZE_FOR_THREAD);
+            let task = tasks.clone();
+            let register = task_register.clone();
             let iterator_stopper = iterator_stopper.clone();
             let task_executor = task_executor.clone();
 
             // workers
             let handle = thread::spawn(move || {
                 loop {
-                    // check stopper flag
                     if iterator_stopper.load(Ordering::SeqCst) {
                         break;
                     }
-                    //  fetch task and register thread number
-                    match get_task(&task_receiver, &task_registry, thread_number) {
-                        // stop if no more task
+                    match get_task(&task, &register, thread_number) {
+                        // finish
                         None => break,
                         Some(task) => match task_executor(task) {
                             Ok(blk) => {
-                                // send output
-                                output_sender.send(blk).unwrap();
+                                sender.send(blk).unwrap();
                             }
                             Err(_) => {
-                                // stop other thread when Error is returned
                                 iterator_stopper.fetch_or(true, Ordering::SeqCst);
                                 break;
                             }
@@ -474,34 +346,36 @@ where
                     }
                 }
             });
-            output_receivers.push(output_receiver);
+            receivers.push(receiver);
             handles.push(handle);
         }
-        handles.push(sender_thread);
 
         ParIterSync {
-            output_receivers,
-            task_registry,
+            receivers,
+            task_order,
             worker_thread: Some(handles),
             iterator_stopper,
             is_killed: false,
-            current: 0,
         }
     }
 }
 
 impl<R> ParIterSync<R> {
     ///
-    /// - stop workers from fetching new tasks
-    /// - pull one result from each worker to prevent `send` blocking
+    /// stop workers from fetching new tasks, and flush remaining works
+    /// to prevent blocking.
     ///
     pub fn kill(&mut self) {
         if !self.is_killed {
             // stop threads from getting new tasks
             self.iterator_stopper.fetch_or(true, Ordering::SeqCst);
-            // receive one for each channel to prevent blocking
-            for receiver in &self.output_receivers {
-                let _ = receiver.try_recv();
+            // flush the remaining tasks in the channel
+            loop {
+                let _ = match self.task_order.recv() {
+                    Ok(thread_number) => self.receivers.get(thread_number).unwrap().recv(),
+                    // all workers have stopped
+                    Err(_) => break,
+                };
             }
             // loop break only when task_order is dropped (all workers have stopped)
             self.is_killed = true;
@@ -510,29 +384,30 @@ impl<R> ParIterSync<R> {
 }
 
 ///
-/// A helper function to receive task from task receiver.
-/// - it also registers thread ID into task registry immediately.
-///
-/// It guarantees to return None if and only if there is no more new task.
+/// A helper function that locks tasks,
+/// register thread_number and task_number
+/// before releasing tasks lock.
 ///
 #[inline(always)]
-fn get_task<T>(
-    tasks: &Receiver<(T, usize)>,
-    registry: &TaskRegistry,
-    thread_number: isize,
+fn get_task<T, TL>(
+    tasks: &Arc<Mutex<TL>>,
+    register: &Sender<usize>,
+    thread_number: usize,
 ) -> Option<T>
 where
     T: Send,
+    TL: Iterator<Item = T>,
 {
     // lock task list
-    // let mut task = tasks.lock().unwrap();
-    // registry task stealing
-    match tasks.recv() {
-        Ok((task, task_id)) => {
-            registry.register(task_id, thread_number);
+    let mut task = tasks.lock().unwrap();
+    let next_task = task.next();
+    // register task stealing
+    match next_task {
+        Some(task) => {
+            register.send(thread_number).unwrap();
             Some(task)
         }
-        Err(_) => None,
+        None => None,
     }
 }
 
@@ -546,15 +421,10 @@ impl<R> Iterator for ParIterSync<R> {
         if self.is_killed {
             return None;
         }
-
-        // look up which thread to fetch result from
-        match self.task_registry.lookup(self.current) {
-            // no more task
-            None => None,
-            Some(thread_num) => {
-                match self.output_receivers[thread_num as usize].recv() {
+        match self.task_order.recv() {
+            Ok(thread_number) => {
+                match self.receivers.get(thread_number).unwrap().recv() {
                     Ok(block) => {
-                        self.current += 1;
                         Some(block)
                     }
                     // some worker have stopped
@@ -564,6 +434,8 @@ impl<R> Iterator for ParIterSync<R> {
                     }
                 }
             }
+            // all workers have stopped
+            Err(_) => None,
         }
     }
 }
@@ -703,9 +575,7 @@ mod test_par_iter {
 
             let results: Vec<i32> = (0..resource_captured.len())
                 .into_par_iter_sync(move |a| error_at_1000(&resource_captured_1, a as i32))
-                .into_par_iter_sync(move |a| {
-                    Ok(resource_captured.get(a as usize).unwrap().to_owned())
-                })
+                .into_par_iter_sync(move |a| Ok(resource_captured.get(a as usize).unwrap().to_owned()))
                 .into_par_iter_sync(move |a| {
                     Ok(resource_captured_2.get(a as usize).unwrap().to_owned())
                 })
